# Copyright 2020 MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.networks.blocks import CRF
from tests.utils import skip_if_no_cpp_extension

TEST_CASES = [
    [
        # Case Description
        "2 batche(s), 1 dimension(s), 2 classe(s), 1 channel(s)",
        # Parameters
        [
            1.0,  # bilateral_weight
            0.3,  # gaussian_weight
            5.0,  # bilateral_spatial_sigma
            0.5,  # bilateral_color_sigma
            5.0,  # gaussian_spatial_sigma
            1.0,  # update_factor
            1,  # compatibility_kernel_range
            5,  # iterations
        ],
        # Input
        [
            # Batch 0
            [
                # Class 0
                [0.8, 0.9, 0.6, 0.2, 0.3],
                # Class 1
                [0.1, 0.3, 0.5, 0.8, 0.7],
            ],
            # Batch 1
            [
                # Class 0
                [0.8, 0.9, 0.6, 0.2, 0.3],
                # Class 1
                [0.1, 0.3, 0.5, 0.8, 0.7],
            ],
        ],
        # Features
        [
            # Batch 0
            [
                # Channel 0
                [1, 1, 1, 0.5, 0],
            ],
            # Batch 1
            [
                # Channel 0
                [1, 1, 0.5, 0, 0],
            ],
        ],
        # Expected
        [
            # Batch 0
            [
                # Class 0
                [0.976472, 0.973789, 0.951958, 0.882982, 0.876651],
                # Class 1
                [0.023528, 0.026211, 0.048042, 0.117018, 0.123349],
            ],
            # Batch 1
            [
                # Class 0
                [0.963642, 0.946892, 0.858650, 0.633639, 0.617334],
                # Class 1
                [0.036358, 0.053108, 0.141350, 0.366361, 0.382666],
            ],
        ],
    ],
    [
        # Case Description
        "1 batche(s), 2 dimension(s), 3 classe(s), 2 channel(s)",
        # Parameters
        [
            1.0,  # bilateral_weight
            0.3,  # gaussian_weight
            5.0,  # bilateral_spatial_sigma
            0.5,  # bilateral_color_sigma
            5.0,  # gaussian_spatial_sigma
            1.0,  # update_factor
            1,  # compatibility_kernel_range
            5,  # iterations
        ],
        # Input
        [
            # Batch 0
            [
                # Class 0
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 1.0],
                    [0.0, 0.0, 0.0, 1.0, 1.0],
                ],
                # Class 1
                [
                    [1.0, 1.0, 0.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ],
                # Class 2
                [
                    [0.0, 0.0, 0.0, 1.0, 1.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0],
                    [0.0, 1.0, 1.0, 1.0, 0.0],
                    [1.0, 1.0, 1.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0, 0.0, 0.0],
                ],
            ],
        ],
        # Features
        [
            # Batch 0
            [
                # Channel 0
                [
                    [1.0, 1.0, 0.0, 0.0, 0.0],
                    [1.0, 1.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                ],
                # Channel 1
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 1.0, 1.0],
                    [0.0, 0.0, 0.0, 1.0, 1.0],
                ],
            ],
        ],
        # Expected
        [
            # Batch 0
            [
                # Class 0
                [
                    [0.000008, 0.000008, 0.000008, 0.000003, 0.000003],
                    [0.000008, 0.000008, 0.000003, 0.000003, 0.000003],
                    [0.000008, 0.000003, 0.000003, 0.000003, 0.000008],
                    [0.000003, 0.000003, 0.000003, 0.000023, 0.000023],
                    [0.000003, 0.000003, 0.000008, 0.000023, 0.000023],
                ],
                # Class 1
                [
                    [0.000023, 0.000023, 0.000008, 0.000003, 0.000003],
                    [0.000023, 0.000023, 0.000003, 0.000003, 0.000003],
                    [0.000008, 0.000003, 0.000003, 0.000003, 0.000008],
                    [0.000003, 0.000003, 0.000003, 0.000008, 0.000008],
                    [0.000003, 0.000003, 0.000008, 0.000008, 0.000008],
                ],
                # Class 2
                [
                    [0.999969, 0.999969, 0.999983, 0.999994, 0.999994],
                    [0.999969, 0.999969, 0.999994, 0.999994, 0.999994],
                    [0.999983, 0.999994, 0.999994, 0.999994, 0.999983],
                    [0.999994, 0.999994, 0.999994, 0.999969, 0.999969],
                    [0.999994, 0.999994, 0.999983, 0.999969, 0.999969],
                ],
            ],
        ],
    ],
    [
        # Case Description
        "1 batche(s), 3 dimension(s), 2 classe(s), 1 channel(s)",
        # Parameters
        [
            1.0,  # bilateral_weight
            0.3,  # gaussian_weight
            5.0,  # bilateral_spatial_sigma
            0.1,  # bilateral_color_sigma
            5.0,  # gaussian_spatial_sigma
            1.0,  # update_factor
            1,  # compatibility_kernel_range
            2,  # iterations
        ],
        # Input
        [
            # Batch 0
            [
                # Class 0
                [
                    # Slice 0
                    [
                        [1.0, 1.0, 0.0, 0.0, 0.0],
                        [1.0, 1.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 1
                    [
                        [1.0, 1.0, 0.0, 0.0, 0.0],
                        [1.0, 1.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 2
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 3
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 4
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                ],
                # Class 1
                [
                    # Slice 0
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 1
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 2
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 3
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0, 1.0],
                        [0.0, 0.0, 0.0, 1.0, 1.0],
                    ],
                    # Slice 4
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0, 1.0],
                        [0.0, 0.0, 0.0, 1.0, 1.0],
                    ],
                ],
            ],
        ],
        # Features
        [
            # Batch 0
            [
                # Channel 0
                [
                    # Slice 0
                    [
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 1
                    [
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                    ],
                    # Slice 2
                    [
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.5, 0.5, 0.5, 0.0, 0.0],
                        [0.5, 0.5, 0.8, 1.0, 1.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                    ],
                    # Slice 3
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                    ],
                    # Slice 4
                    [
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                        [0.0, 0.0, 1.0, 1.0, 1.0],
                    ],
                ],
            ],
        ],
        # Expected
        [
            # Batch 0
            [
                # Class 0
                [
                    # Slice 0
                    [
                        [1.000000, 1.000000, 1.000000, 0.999808, 0.721122],
                        [1.000000, 1.000000, 1.000000, 0.999758, 0.666025],
                        [1.000000, 1.000000, 0.999979, 0.995894, 0.577459],
                        [0.999787, 0.999739, 0.995725, 0.934170, 0.488704],
                        [0.691645, 0.641028, 0.560127, 0.481718, 0.431180],
                    ],
                    # Slice 1
                    [
                        [1.000000, 1.000000, 1.000000, 0.999743, 0.650416],
                        [1.000000, 1.000000, 0.999999, 0.992747, 0.108034],
                        [1.000000, 0.999999, 0.998541, 0.402370, 0.007122],
                        [0.999711, 0.992109, 0.391941, 0.003358, 0.000440],
                        [0.615523, 0.097120, 0.006599, 0.000427, 0.000365],
                    ],
                    # Slice 2
                    [
                        [1.000000, 1.000000, 0.999975, 0.995241, 0.543122],
                        [1.000000, 0.999998, 0.998394, 0.381981, 0.006586],
                        [0.999973, 0.998313, 0.370238, 0.000596, 0.000034],
                        [0.994611, 0.361317, 0.000573, 0.000001, 0.000000],
                        [0.505392, 0.005862, 0.000032, 0.000000, 0.000000],
                    ],
                    # Slice 3
                    [
                        [0.999692, 0.999639, 0.994364, 0.919713, 0.446683],
                        [0.999626, 0.990123, 0.347190, 0.002895, 0.000390],
                        [0.993872, 0.336665, 0.000525, 0.000001, 0.000000],
                        [0.910704, 0.002676, 0.000001, 0.000000, 0.000000],
                        [0.413964, 0.000354, 0.000000, 0.000000, 0.000000],
                    ],
                    # Slice 4
                    [
                        [0.574496, 0.533306, 0.469335, 0.419446, 0.390403],
                        [0.524093, 0.072180, 0.005087, 0.000354, 0.000318],
                        [0.449362, 0.004875, 0.000028, 0.000000, 0.000000],
                        [0.393431, 0.000331, 0.000000, 0.000000, 0.000000],
                        [0.362417, 0.000295, 0.000000, 0.000000, 0.000000],
                    ],
                ],
                # Class 1
                [
                    # Slice 0
                    [
                        [0.000000, 0.000000, 0.000000, 0.000192, 0.278878],
                        [0.000000, 0.000000, 0.000000, 0.000242, 0.333975],
                        [0.000000, 0.000000, 0.000021, 0.004106, 0.422541],
                        [0.000213, 0.000261, 0.004275, 0.065830, 0.511296],
                        [0.308355, 0.358972, 0.439873, 0.518282, 0.568820],
                    ],
                    # Slice 1
                    [
                        [0.000000, 0.000000, 0.000000, 0.000257, 0.349584],
                        [0.000000, 0.000000, 0.000001, 0.007253, 0.891966],
                        [0.000000, 0.000001, 0.001459, 0.597630, 0.992878],
                        [0.000289, 0.007891, 0.608059, 0.996642, 0.999560],
                        [0.384477, 0.902880, 0.993401, 0.999573, 0.999635],
                    ],
                    # Slice 2
                    [
                        [0.000000, 0.000000, 0.000025, 0.004759, 0.456878],
                        [0.000000, 0.000002, 0.001606, 0.618019, 0.993414],
                        [0.000027, 0.001687, 0.629762, 0.999404, 0.999966],
                        [0.005389, 0.638683, 0.999427, 0.999999, 1.000000],
                        [0.494608, 0.994138, 0.999968, 1.000000, 1.000000],
                    ],
                    # Slice 3
                    [
                        [0.000308, 0.000361, 0.005636, 0.080287, 0.553317],
                        [0.000374, 0.009877, 0.652810, 0.997105, 0.999610],
                        [0.006128, 0.663335, 0.999475, 0.999999, 1.000000],
                        [0.089296, 0.997324, 0.999999, 1.000000, 1.000000],
                        [0.586036, 0.999646, 1.000000, 1.000000, 1.000000],
                    ],
                    # Slice 4
                    [
                        [0.425504, 0.466694, 0.530665, 0.580554, 0.609597],
                        [0.475907, 0.927820, 0.994913, 0.999646, 0.999682],
                        [0.550638, 0.995125, 0.999972, 1.000000, 1.000000],
                        [0.606569, 0.999669, 1.000000, 1.000000, 1.000000],
                        [0.637583, 0.999705, 1.000000, 1.000000, 1.000000],
                    ],
                ],
            ],
        ],
    ],
]


@skip_if_no_cpp_extension
class CRFTestCaseCpu(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test(self, test_case_description, params, input, features, expected):

        # Create input tensors
        input_tensor = torch.from_numpy(np.array(input)).to(dtype=torch.float, device=torch.device("cpu"))
        feature_tensor = torch.from_numpy(np.array(features)).to(dtype=torch.float, device=torch.device("cpu"))

        # apply filter
        crf = CRF(*params)
        output = crf(input_tensor, feature_tensor).cpu().numpy()

        # Ensure result are as expected
        np.testing.assert_allclose(output, expected, atol=1e-4)


if __name__ == "__main__":
    unittest.main()
